"""Distributional plots."""

import numpy as np
import pandas as pd
from seaborn._statistics import Histogram

from ezcharts.plots import _HistogramPlot, util


__all__ = ["displot", "histplot", "kdeplot", "ecdfplot", "rugplot", "distplot"]


def displot(*args, **kwargs):
    """Figure-level interface for distribution plots in a grid."""
    raise NotImplementedError


class MakeRectangles(util.JSCode):
    """
    Make JSCode rectangles.

    Use Instance as an argument to series.renderItem.
    Series.encode should list two x dimensions for bar start and end
    and one y dimension for bar height.
    """

    def __init__(self):
        """Instantiate the class with some jscode.

        Each call to this function creates JS code to make a single echarts
        rectangle.

        First get the x-start, x-end and height of the rectangle in raw data
        coords. eg: var data_x_start = api.value(0);

        Convert the raw data coords to canvas coords.
        Note in canvas coords (x=0, y=0) is in the top left.
        eg: var start = api.coord([data_start, 0]);

        start[1] is y=0 in canvas coords. Subtract the height from this to get
        the canvas coords for the top of the rectangle.
        """
        jscode = """function renderItem(params, api) {
            var data_x_start = api.value(0);
            var data_x_end = api.value(1);
            var data_height = api.value(2);

            var start = api.coord([data_x_start, 0]);
            var end = api.coord([data_x_end, 0]);
            var height = api.size([0, data_height])[1];

            var rectShape = echarts.graphic.clipRectByRect(
                {
                    x: start[0],
                    y: start[1] - height,
                    width: end[0] - start[0],
                    height: height
                },
                {
                    x: params.coordSys.x,
                    y: params.coordSys.y,
                    width: params.coordSys.width,
                    height: params.coordSys.height
                }
            );
            return (
                rectShape && {
                    type: 'rect',
                    transition: ['shape'],
                    shape: rectShape,
                    style: api.style()

                }
            );}"""

        super().__init__(jscode)


def histplot(
    data=None, *, x=None, y=None, hue=None, weights=None,
    stat='count', bins='auto', binwidth=None, binrange=None,
    discrete=None, cumulative=False, common_bins=True,
    common_norm=True, multiple='layer', element='bars',
    fill=True, shrink=1, kde=False, kde_kws=None,
    line_kws=None, thresh=0, pthresh=None, pmax=None,
    cbar=False, cbar_ax=None, cbar_kws=None, palette=None,
    hue_order=None, hue_norm=None, color=None, log_scale=None,
        legend=True, ax=None, **kwargs):
    """Plot univariate or multivariate histograms."""
    plt = _HistogramPlot()
    plt.xAxis = dict()
    plt.yAxis = dict()

    estimate_kws = dict(
        stat=stat,
        bins=bins,
        binwidth=binwidth,
        binrange=binrange,
        discrete=discrete,
        cumulative=cumulative,
    )

    data = pd.DataFrame(data)

    if data.shape[1] > 1:
        # multivariate data
        opacity = 0.5
        if palette is None:
            palette = util.choose_palette(ncolours=data.shape[1])
    else:
        opacity = 1.0
        if color is None:
            palette = util.choose_palette(ncolours=1)
        else:
            palette = [color]

    if hue:
        data = data.pivot(columns=hue, values=data.columns[0])

    for dataset_idx, col in enumerate(data.columns):
        variable_data = data[col].dropna()
        estimator = Histogram(**estimate_kws)
        heights, edges = estimator(variable_data, weights=weights)

        x_starts = edges[:-1]
        x_ends = edges[1:]
        rect_data = np.stack((x_starts, x_ends, heights), axis=-1)

        plt.add_dataset(dict(
            source=rect_data,
            dimensions=['x_start', 'x_end', 'height']))

        plt.add_series(dict(
            name=str(col),
            type='custom',
            renderItem=MakeRectangles(),
            itemStyle=dict(opacity=opacity, color=palette[dataset_idx]),
            datasetIndex=dataset_idx,
            encode={
                'x': ['x_start', 'x_end'],
                'y': ['height']
            },
            clip=True
            ))
        plt.xAxis.type = 'value'
    return plt


def kdeplot(*args, **kwargs):
    """Plot uni/bi-variate distributions using kernel density estimation."""
    raise NotImplementedError


def ecdfplot(*args, **kwargs):
    """Plot empirical cumulative distribution functions."""
    raise NotImplementedError


def rugplot(*args, **kwargs):
    """Plot marginal distributions by drawing ticks along the x and y axes."""
    raise NotImplementedError


def distplot(*args, **kwargs):
    """Flexibly plot a univariate distribution of observations (DEPRECATED)."""
    raise DeprecationWarning
    raise NotImplementedError



"""Get default cards layouts."""
from typing import Optional, Type
from uuid import uuid4

from dominate.tags import button, div, h5, html_tag

from ezcharts.layout.base import IClasses, Snippet
from ezcharts.layout.util import cls


class IOffCanvasClasses(IClasses):
    """Cards html classes."""

    container: str = cls("container", "px-0")
    offcanvas: str = cls("offcanvas", "offcanvas-bottom", "h-auto")
    offcanvas_button: str = cls("")
    offcanvas_header: str = cls("offcanvas-header")
    offcanvas_title: str = cls("offcanvas-title")
    offcanvas_body: str = cls("offcanvas-body")


class OffCanvas(Snippet):
    """A styled section tag."""

    TAG = 'div'

    def __init__(
        self,
        title,
        label,
        body,
        classes: IOffCanvasClasses = IOffCanvasClasses(),
    ) -> None:
        """Create the layout."""
        super().__init__(
            styles=None,
            classes=classes,
            className=classes.container)

        with self:
            self.add_offcanvas(
                title,
                label,
                f"OffCanvas_{str(uuid4()).replace('-', '')}",
                body)

    def add_offcanvas(
        self,
        title: str,
        label: str,
        new_uid: str,
        body: Optional[str] = None,
        offcanvas_cls: Optional[str] = None,
        offcanvas_btn_cls: Optional[str] = "btn-info",
        head_tag: Type[html_tag] = h5
    ) -> None:
        """Add cards to the grid."""
        button(
                label,
                type="button",
                data_bs_toggle="offcanvas",
                data_bs_target=f"#{new_uid}",
                aria_controls="offcanvasBottom",
                className=cls(
                    self.classes.offcanvas_button,
                    offcanvas_btn_cls if offcanvas_btn_cls is not None else '')
                )

        with div(
            id=new_uid,
            tabindex=-1,
            aria_labelledby="offcanvasBottomLabel",
            className=cls(
                self.classes.offcanvas,
                offcanvas_cls if offcanvas_cls is not None else '')):

            with div(className=cls(
                    self.classes.offcanvas_header,
                    'container',
                    'px-0')):
                head_tag(title, className=self.classes.offcanvas_title)
                button(
                    type="button",
                    cls="btn-close",
                    data_bs_dismiss="offcanvas",
                    aria_label="Close")

            div(body, className=self.classes.offcanvas_body)



"""Get default summary bar layouts."""
from dominate.tags import div, h1, p

from ezcharts.layout.base import IClasses, IStyles, Snippet
from ezcharts.layout.util import cls, css


class IBadgeClasses(IClasses):
    """Section html classes."""

    container: str = cls(
        "badge", "px-3", "py-2", "mb-3", "me-3", "badge")
    container_bg: str = cls("bg-primary")


class IBadgeStyles(IStyles):
    """Section inline css styles."""

    container: str = css(
        "line-height: 20px", "border-radius: 6px",
        "font-size: 13px;")


class Badge(Snippet):
    """A styled span."""

    TAG = 'span'

    def __init__(
        self,
        title: str,
        bg_class=None,
        styles: IBadgeStyles = IBadgeStyles(),
        classes: IBadgeClasses = IBadgeClasses(),
    ) -> None:
        """Create styled badge."""
        bg = bg_class or classes.container_bg
        super().__init__(
            title,
            styles=styles,
            classes=classes,
            className=f"{classes.container} {bg}",
            style=styles.container)


class IBannerClasses(IClasses):
    """Section html classes."""

    container: str = cls("px-4", "bg-dark", "labs-banner")
    inner: str = cls(
        "container", "px-0", "py-3", "border-top", "text-white", "report-title")


class IBannerStyles(IStyles):
    """Section inline css styles."""

    container: str = css(
        "margin-bottom: -25px",
        "padding-bottom: 35px !important")
    inner: str = css("border-color: rgba(255, 255, 255, 0.1) !important;")


class Banner(Snippet):
    """A styled div tag containing a heading and badges."""

    TAG = 'div'

    def __init__(
        self,
        report_title: str,
        workflow_name: str,
        styles: IBannerStyles = IBannerStyles(),
        classes: IBannerClasses = IBannerClasses(),
        default_content: bool = True
    ) -> None:
        """Create styled banner."""
        super().__init__(
            styles=styles,
            classes=classes,
            className=classes.container,
            style=styles.container)

        with self:
            with div(className=classes.inner, style=styles.inner):
                if not default_content:
                    return
                h1(report_title)
                p(
                    f"Results generated through the {workflow_name} nextflow "
                    "workflow provided by Oxford Nanopore Technologies.",
                    className="py-3 fs-5")
                self.badges = div(className="d-flex flex-wrap")

    def add_badge(
        self,
        title: str,
        bg_class=None,
        styles: IBadgeStyles = IBadgeStyles(),
        classes: IBadgeClasses = IBadgeClasses(),
    ) -> None:
        """Add a badge to the banner."""
        with self.badges:
            Badge(title, styles=styles, classes=classes, bg_class=bg_class)



"""An ezcharts component for plotting sequence summaries."""
import argparse

import numpy as np
import pandas as pd
from pkg_resources import resource_filename

import ezcharts as ezc
from ezcharts.components.ezchart import EZChart
from ezcharts.components.reports.comp import ComponentReport
from ezcharts.layout.base import Snippet
from ezcharts.layout.snippets import Grid, Tabs
from ezcharts.plots import util


class SeqSummary(Snippet):
    """Generate sequence summary plots."""

    def __init__(self, seq_summary, theme='epi2melabs', **kwargs):
        """Create sequence summary componet.

        If seq_summary contains results from multiple samples, each will be
        plotted in its own tab.

        :param seq_summary: A path to a fastcat / bamstats output file or
            DataFrame
        """
        super().__init__(styles=None, classes=None)

        with self:
            if not isinstance(seq_summary, pd.DataFrame):
                df_all = util.read_files(seq_summary)
            else:
                df_all = seq_summary
            # Out of the intersection of columns between `fastcat` and
            # `bamstats` output, only the column with the read IDs has a
            # different name (`read_id` in `fastcat` and `name` in `bamstats`).
            # Rename to ensure the dataframes produced from either `fastcat` or
            # `bamstats` are consistent.
            if df_all.columns[0] == "name":
                df_all.rename(columns={"name": "read_id"}, inplace=True)
            # the following assumes that `fastcat` / `bamstats` has been run
            # with `-s` which is not necessarily the case (if there was only
            # one sample) --> create a dummy 'sample_name' column if missing
            if 'sample_name' not in df_all.columns:
                df_all['sample_name'] = 'sample'
            if len(df_all['sample_name'].unique()) == 1:
                # we only got a single sample --> no dropdown
                draw_all_plots(df_all, theme)
            else:
                # several samples --> use a dropdown menu
                tabs = Tabs()
                with tabs.add_dropdown_menu():
                    for sample_id, df_sample in df_all.groupby('sample_name'):
                        with tabs.add_dropdown_tab(sample_id):
                            draw_all_plots(df_sample, theme)


def draw_all_plots(seq_summary, theme):
    """Draw all three plots."""
    with Grid(columns=3):
        EZChart(read_quality_plot(seq_summary), theme)
        EZChart(read_length_plot(seq_summary), theme)
        EZChart(base_yield_plot(seq_summary), theme)


def base_yield_plot(
        seq_summary,
        title='Base yield above read length'):
    """Create base yield plot."""
    if not isinstance(seq_summary, pd.DataFrame):
        df = util.read_files(seq_summary)[['read_length']]
    else:
        df = seq_summary
    df = df.sort_values('read_length', ascending=True)
    df = pd.concat(
        (pd.DataFrame.from_dict({'read_length': 0}, orient='index').T, df))

    ylab = 'Yield above length / Gbases'
    xlab = 'Read length / kb'

    # If we have u/int8 or u/int16 cast to float to prevent overflow
    df.read_length = df.read_length.astype('uint64')
    df[ylab] = \
        df.read_length.cumsum()[::-1].values / 1e+9
    df[xlab] = df['read_length'] / 1000

    # No need plot all the points
    if len(df) > 1000:
        step = len(df) // 1000
        # thin the data while keeping the last data point
        df = pd.concat((df.loc[::step, :], df.iloc[[-1]]), axis=0)

    plt = ezc.lineplot(data=df, x=xlab, y=ylab, hue=None)
    plt.series[0].showSymbol = False
    plt.title = dict(text=title)
    return plt


def read_quality_plot(
        seq_summary, binwidth=0.2,
        min_qual=4, max_qual=30, title='Read quality'):
    """Create read quality summary plot."""
    if not isinstance(seq_summary, pd.DataFrame):
        df = util.read_files(seq_summary)[['mean_quality']]
    else:
        df = seq_summary
    mean_q = np.round(df.mean_quality.mean(), 1)
    median_q = int(df.mean_quality.median())

    plt = ezc.histplot(
        data=df.mean_quality, binwidth=binwidth, binrange=(min_qual, max_qual))
    plt.title = dict(
        text=title,
        subtext=f"Mean: {mean_q}. Median: {median_q}")
    plt.xAxis.name = 'Quality score'
    plt.xAxis.min, plt.xAxis.max = min_qual, max_qual
    plt.yAxis.name = 'Number of reads'
    return plt


def read_length_plot(
        seq_summary, min_len=None, max_len=None,
        xlim=(0, None), bins=100, bin_width=None, title='Read length'):
    """Create a read length plot.

    :param seq_summary: summary data from fastcat.
    :param min_len: minimum length.
    :param max_len: maximum length.
    :param xlim: tuple for plotting limits (start, end). A value None will
        trigger calculation from the data.
    :param: bins number of bins

    The minimum and maximum lengths are used only to annotate the plot
    (not filter the data).
    """
    if not isinstance(seq_summary, pd.DataFrame):
        df = util.read_files(seq_summary)[['read_length']]
    else:
        df = seq_summary

    mean_length = int(df['read_length'].mean())
    median_length = int(np.median(df['read_length']))
    max_ = int(np.max(df['read_length']))

    plt = ezc.histplot(data=df.read_length / 1000, bins=bins)
    plt.title = dict(
        text=title,
        subtext=f"Mean: {mean_length}. Median: {median_length}. Max: {max_}")
    plt.xAxis.name = 'Read length / kb'
    plt.yAxis.name = 'Number of reads'
    return plt


def main(args):
    """Entry point to demonstrate a sequence summmary component."""
    comp_title = 'Sequence Summary'
    seq_sum = SeqSummary(args.seq_summary)
    report = ComponentReport(comp_title, seq_sum)
    report.write(args.output)


def argparser():
    """Argument parser for entrypoint."""
    parser = argparse.ArgumentParser(
        'Sequence summary',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=False)
    parser.add_argument(
        "--seq_summary",
        default=resource_filename('ezcharts', "data/test/fastcat.stats.gz"),
        help="Sequence summary TSV from fastcat.")
    parser.add_argument(
        "--output",
        default="seq_summary_report.html",
        help="Output HTML file.")
    return parser



"""Nextclade reporting section."""

import argparse

from dominate.tags import script
from dominate.util import raw

from ezcharts.components.reports.comp import ComponentReport
from ezcharts.components.theme import LAB_head_resources
from ezcharts.layout.base import Snippet
from ezcharts.layout.resource import ScriptResource
from ezcharts.layout.util import inline, render_template


class NextClade(Snippet):
    """A nextclade report component."""

    TAG = 'nxt-table'

    def __init__(
        self,
        json_path
    ) -> None:
        """Initialize a nextclade instance."""
        super().__init__(
            styles=None,
            classes=None)

        with open(json_path, encoding='utf8') as fh:
            data = fh.read()

        with self:
            script(raw(render_template(
                template=(
                    """
                    const data = {{ data }}
                    var nxt = document.querySelector('nxt-table')
                    nxt.data = data.results
                    """),
                data=data)))


NXTComponent = ScriptResource(
    path='nextclade.html',
    loader=inline)


def main(args):
    """Entry point to create a report from nextclade."""
    nxt = NextClade(args.json)
    report = ComponentReport(
        'Nextclade', nxt,
        head_resources=[*LAB_head_resources, NXTComponent])
    report.write(args.output)


def argparser():
    """Argument parser for entrypoint."""
    parser = argparse.ArgumentParser(
        'Params table',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=False
    )
    parser.add_argument(
        "json",
        help="JSON output file from nextclade CLI."
    )
    parser.add_argument(
        "--output",
        default="nextclade_report.html",
        help="Output HTML file."
    )
    return parser



"""Useful reusable functions."""
import contextlib
import json
import os
from typing import Dict
import warnings

from dominate.tags import script, style
from dominate.util import container, raw, text
from jinja2 import BaseLoader, Environment
from pkg_resources import resource_filename
import sass

warnings.simplefilter("always", DeprecationWarning)


def load_json(
    path: str
) -> Dict:
    """Return json file as a dict."""
    with open(path, 'r', encoding='utf-8') as content:
        return json.load(content)


def inline(
    path: str
) -> text:
    """Return a file as a string."""
    with open(path, 'r', encoding='utf-8') as content:
        return raw(content.read())


def resolve_import(path):
    """Find resource and return as list."""
    resolved = resource_filename('ezcharts', path)
    # Temporary measure
    if os.path.exists(resolved) and not os.path.isdir(resolved):
        return [[resolved]]
    return [[path]]


def transpile(path):
    """Compile scss to css."""
    compiled = sass.compile(
        filename=path,
        output_style='compressed',
        importers=[(0, resolve_import)])
    return raw(compiled)


def inline_script(path):
    """Inline a script from path."""
    return script(inline(path))


def transpile_style(path):
    """Transpile a stylesheet from path."""
    return style(transpile(path))


def write_report(path, document):
    """Write a report to file."""
    with open(path, 'w', encoding='utf-8') as out:
        out.write('<!DOCTYPE html>')
        out.write(document.render())


def render_template(template, **kwargs):
    """Render a jinja2 template."""
    rtemplate = Environment(
        loader=BaseLoader()).from_string(template)
    return rtemplate.render(**kwargs)


def cls(*classes: str) -> str:
    """Collect element classes from a list."""
    return ' '.join(classes)


def css(*styles: str) -> str:
    """Collect inline element styles from a list."""
    return ' '.join(s.strip(';') + ';' for s in styles)


@contextlib.contextmanager
def isolate_context():
    """Prevents `dominate` from adding tags to the enclosing context.

    Some ezCharts `Snippets` use dominate context managers in their constructors.
    Therefore, they are automatically added to the enclosing `dominate` context when
    created, which can be undesired in some cases and can be prevented with this
    function.
    """
    warnings.warn(
        "`layout.util.isolate_context()` will become redundant (and subsequently "
        "removed) once `ezCharts` makes the switch to use `dominate`s declarative "
        "syntax for the report components.",
        DeprecationWarning,
        stacklevel=3,
    )
    # `dominate.util.container` is a tag which can contain elements, but does not add an
    # extra level
    with container() as c:
        yield
    # remove everything that has been added to `c`
    c.clear()



"""Re-usable external resources."""
from dataclasses import dataclass
from typing import Callable, Optional, Type

from dominate.tags import dom_tag, script, style
from pkg_resources import resource_filename

from ezcharts.layout.util import inline


@dataclass
class Resource:
    """Resource definition."""

    path: str
    loader: Callable
    tag: Optional[Type[dom_tag]] = None

    @property
    def data_file(self):
        """Filepath to resource data file."""
        return resource_filename('ezcharts', f'data/{self.path}')

    def __call__(self):
        """Render the resource."""
        loaded = self.loader(self.data_file)
        if self.tag:
            return self.tag(loaded)
        return loaded


# dataclasses are annoying to derive


def VendorResource(path, loader=inline, tag=None):
    """Fetch a vendor resource."""
    return Resource(f'vendor/{path}', loader, tag)


def ScriptResource(path, loader=inline, tag=None):
    """Fetch a script resource."""
    return Resource(f'scripts/{path}', loader, tag)


def ImageResource(path, loader=inline, tag=None):
    """Fetch an image resource."""
    return Resource(f'images/{path}', loader, tag)


def StyleResource(path, loader=inline, tag=None):
    """Fetch a style resource."""
    return Resource(f'styles/{path}', loader, tag)


def ThemeResource(path, loader=inline, tag=None):
    """Fetch a theme resource."""
    return Resource(f'themes/{path}', loader, tag)


echarts_js = VendorResource(
    path='echarts.min.js',
    tag=script,
    loader=inline)

bootstrap_js = VendorResource(
    path='bootstrap-5.3.0/js/bootstrap.bundle.min.js',
    tag=script,
    loader=inline)

datatables_js = VendorResource(
    path='simple-datatables/simple-datatables.js',
    tag=script,
    loader=inline)

datatables_css = VendorResource(
    path='simple-datatables/simple-datatables.css',
    tag=style,
    loader=inline)

base_head_resources = [
    echarts_js,
    datatables_js,
    datatables_css]

base_body_resources = [
    bootstrap_js]



"""The odd helper function."""

import argparse
import logging


def get_named_logger(name):
    """Create a logger with a name.

    :param name: name of logger.
    """
    name = name.ljust(10)[:10]  # so logging is aligned
    logger = logging.getLogger('{}.{}'.format(__package__, name))
    logger.name = name
    return logger


def _log_level():
    """Parser to set logging level and acquire software version/commit."""
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter, add_help=False)

    modify_log_level = parser.add_mutually_exclusive_group()
    modify_log_level.add_argument(
        '--debug', action='store_const',
        dest='log_level', const=logging.DEBUG, default=logging.INFO,
        help='Verbose logging of debug information.')
    modify_log_level.add_argument(
        '--quiet', action='store_const',
        dest='log_level', const=logging.WARNING, default=logging.INFO,
        help='Minimal logging; warnings only.')

    return parser



"""Sunburst plots."""

from ezcharts.plots import _NoAxisFixPlot, util


__all__ = ["sunburst"]


def sunburst(
    data, tooltip=True, visualMap=True, min_value=0, max_value=None,
    color_scale=None, label_rotate="radial", label_minAngle=0, show_label=True
):
    """Create sunburst plot.

    :param data (list): List of nested dictionaries of the following structure:
        E.g.: `[{'name': 'parent A', 'value': X+Y, 'children':[
        {'name': 'child A1', 'value': X}, {'name': 'child A2', 'value': Y}
        ] },
        ...
        ]`.
    :param tooltip (bool): Enable/Disable tooltip.
    :param visualMap (bool): Enable/Disable `visualMap`. By default `visualMap` colors
        each section by their `value`. Color gradient is controlled with `min_value` and
        `max_value`.
    :param min_value (int): Minimum value.
    :param max_value (int): Maximum value.
    :param color_scale (list): List of colors to establish a gradient between
        `min_value` and `max_value`.
    :param label_rotate (int, str): If it is number type, then is stands for rotation,
        from -90 degrees to 90 degrees, and positive values stand for counterclockwise.
        It can also be 'radial' (along the radius) or 'tangential' (perpendicular to
        the radius).
    :param label_minAngle (int): Do not display the text if angle of data piece is
        smaller than this value (in degrees).
    :param show_label (bool): Whether to show labels.
    :return (plot): Sunburst plot.
    """
    plt = _NoAxisFixPlot()  # we don't have an axis so no point trying to fix
    plt.xAxis = dict(show=False)
    plt.yAxis = dict(show=False)
    plt.add_series(
        dict(
            type="sunburst",
            data=data,
            radius=[0, "100%"],
            emphasis=dict(focus="ancestor"),
            label=dict(
                rotate=label_rotate, minAngle=label_minAngle, show=show_label),
        )
    )
    # Enable tooltip
    if tooltip:
        plt.tooltip = dict(trigger="item")
    # Enable visualMap to color portions according to their abundance
    if visualMap:
        if max_value is None:
            # There are two options, all the levels can have their own values or just
            # the terminal children nodes. In this case, parent level values are
            # calculated adding the terminal values.
            # Check this and calculate max_value depending on input data.
            # if the first element have its own value:
            if data[0].get('value'):
                max_value = max({i.get("value") for i in data})
            else:
                max_value = sum_terminal_nodes(data)
        if not color_scale:
            color_scale = util.choose_palette('crest')
        plt.visualMap = [
            dict(
                type="continuous",
                min=min_value,
                max=max_value,
                inRange=dict(color=color_scale),
            )
        ]

    return plt


def sum_terminal_nodes(node):
    """Get the sum of the terminal nodes.

    :param node (dictionary): dictionary that contains name, value (optional) and
        and children (optional).
    :return (int): sum of the terminal nodes.
    """
    if "children" in node:
        # recursive case
        total = 0
        for child in node["children"]:
            total += sum_terminal_nodes(child)
        return total
    else:
        # base case
        return node["value"]



"""Get default cards layouts."""
from typing import Optional

from dominate.tags import div, span

from ezcharts.layout.base import IClasses, Snippet
from ezcharts.layout.util import cls


class IProgressClasses(IClasses):
    """Progress html classes."""

    container: str = None
    progress: str = cls("progress align-items-center position-relative")
    progress_bar: str = cls("progress-bar")


class Progress(Snippet):
    """A styled section tag."""

    TAG = 'div'

    def __init__(
        self,
        value_min: int = 0,
        value_max: int = 100,
        value_now: int = 50,
        display_val: Optional[str] = None,
        bar_cls: Optional[str] = None,
        height: Optional[int] = None,
        title: Optional[str] = None,
        styles: Optional[str] = None,
        classes: IProgressClasses = IProgressClasses()
    ) -> None:
        """Create the layout."""
        super().__init__(
            styles=None,
            classes=classes,
            className=classes.container)

        with self:
            self.add_progress(
                value_min,
                value_max,
                value_now,
                display_val,
                bar_cls,
                title,
                height)

    def add_progress(
        self,
        value_min: int,
        value_max: int,
        value_now: int,
        display_val: Optional[str] = None,
        bar_cls: Optional[str] = None,
        title: Optional[str] = None,
        height: Optional[int] = None
    ) -> None:
        """Add progress bar to the layout."""
        # Add a title of one is defined
        span(title, className=cls("fw-bold")) if title is not None else ''

        # Set style if height of bar is defined
        pstyle = f"height: {str(height)}px" if height is not None else ''

        # calculate the width in percent
        width = 100 * (value_now / value_max)

        # define text to be displayed in the progress bar
        display_val = (
            display_val if display_val is not None else f"{width:.1f}%"
        )

        ptext = "justify-content-center d-flex position-absolute w-100"

        with div(className=cls(self.classes.progress), style=pstyle):
            div(
                span(
                    display_val,
                    cls=ptext,
                    style='' if height is None else f"font-size:{height*4}%"),
                role="progressbar",
                style=f"width:{width}%;height:100%;",
                aria_valuenow=str(value_now),
                aria_valuemin=str(value_min),
                aria_valuemax=str(value_max),
                className=cls(
                    self.classes.progress_bar,
                    bar_cls if bar_cls is not None else ''))



"""Mulit-plot grids."""

__all__ = ['pairplot', 'jointplot']


def pairplot(
        *args, **kwargs):
    """Plot pairwise relationships in a dataset."""
    raise NotImplementedError


def jointplot(
        *args, **kwargs):
    """Draw a plot of two variables with bivariate and univariate graphs."""
    raise NotImplementedError
